
--[[
    This file is a library to setup a lua 5.1 restricted environement
    ----

    Copyright (C) 2008-2011  TsT <tst@worldmaster.fr>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
]]--

--[[
  Hello,
  If you found a bug or have a question, don't hesitate to contact me
  by email at <tst@worldmaster.fr> (in french or english please)
  Thanks,
  Have fun!
]]--

--[[ VERSION=0.2.0 20110430 ]]--
local luajailversion = "luajail 0.2.0"

assert(_VERSION and _VERSION == "Lua 5.1",
	"Version mismatch. This script are made for Lua 5.1")

assert(setfenv, "setfenv require")

-- the problem explained here :
--   http://stackoverflow.com/questions/3098544/lua-variable-scoping-with-setfenv
-- require behavior : http://www.lua.org/pil/8.1.html

local function jail_dofile(name)
	local f,e = loadfile(name)
	if not f then error(e, 2) end
	setfenv(f, getfenv(2))
	return f()
end

local function jail_print(...) print("jail:", ...) end


-----------------------------
--[[ lua functions ]]--------
-----------------------------

local jail = {}
jail._VERSION = _G._VERSION -- the lua version
jail._G = jail

jail.jailed = true -- expose that is a jail ! useful for debug


-- getfenv() or getfenv(0) return the out-of-jail _G
local function jail_getfenv(f)
	if type(f) == "function" then
		return getfenv(f)
	end
	return jail._G
end

-- reproduce the require behavior
-- we maybe just need to modify the loaders ?
-- but only if require do not directly modify the out-of-jail package variable
local function jail_require(modname)
	print("jail_require ", modname) -- for debug
	local loaders = jail.package.loaders
	local function find(loader, modname)
		if loader then
			local f = loader(modname)
			if type(f) == "function" then
				setfenv(f, jail)
				return true, f()
			end
		end
		return false
	end
	local function ifFound(modname, found, value)
		if found then
			if value then
				jail.package.loaded[modname] = value
			elseif not package.loaded[modname] then
				jail.package.loaded[modname] = true
			end
			return true
		end
		return false
	end

	for i,loader in ipairs(loaders) do
		if ifFound(modname, find(loader, modname)) then
			break
		end
	end
	-- TODO: handle error cases!
	return jail.package.loaded[modname]
end

jail._VERSION		= _VERSION		-- ok
jail.assert		= assert		-- ok
jail.collectgarbage	= collectgarbage	-- seems ok
jail.dofile		= jail_dofile		-- dangerous : go outside the love filesystem jail
jail.error		= error			-- seems ok
jail.getfenv		= jail_getfenv		-- RISK: lua env jailbreak risk
jail.getmetatable	= getmetatable		-- ok
jail.ipairs		= ipairs		-- ok
jail.load		= load			-- seems ok
--loadfile					-- dangerous : go outside the love filesystem jail
jail.loadstring		= loadstring		-- seems ok
jail.module		= module		-- remove it ?
jail.next		= next			-- ok
jail.pairs		= pairs			-- ok
jail.pcall		= pcall			-- ok
jail.print		= jail_print		-- ok (for debug)
jail.rawequal		= rawequal		-- ok
jail.rawget		= rawget		-- ok
jail.rawset		= rawset		-- ok
jail.require		= jail_require		-- RISK
jail.select		= select		-- ok
--setfenv		= setfenv		-- seems not ok, is able to reset the current env ?
jail.setmetatable	= setmetatable		-- ok
jail.tonumber		= tonumber		-- ok
jail.tostring		= tostring		-- ok
jail.type		= type			-- ok
jail.unpack		= unpack		-- ok
jail.xpcall		= xpcall		-- ok


-----------------------------
--[[ os ]]-------------------
-----------------------------

jail.os = {}
jail.os.clock     = os.clock		-- ok
jail.os.date      = os.date		-- ok
jail.os.difftime  = os.difftime		-- ok
jail.os.getenv    = os.getenv		-- ok
jail.os.setlocale = os.setlocale	-- seems ok
jail.os.time      = os.time		-- ok
--jail.os.execute   = os.execute	-- RISK
--jail.os.exit      = os.exit		-- ?
--jail.os.remove    = os.remove		-- RISK
--jail.os.rename    = os.rename		-- RISK
--jail.os.tmpname   = os.tmpname	-- RISK


-----------------------------
--[[ io ]]-------------------
-----------------------------

jail.io = {}
jail.io.flush        = io.flush		-- seems ok
-- RISK: file loading jailbreak risk	-- patched, ok
jail.io.input = function() -- no file argument
        return io.input()
end

-- RISK: file writing jailbreak risk	-- patched, ok
jail.io.output       = function() -- no file argument
        return io.output()
end
jail.io.read         = io.read -- really ?
jail.io.stderr       = io.stderr	-- ok
jail.io.stdin        = io.stdin		-- ok
jail.io.stdout       = io.stdout	-- ok
jail.io.type         = io.type		-- ok
jail.io.write        = io.write		-- seems ok (is for stdout, not file)
--jail.io.open         = io.open	-- denied by love policy
--jail.io.close        = io.close	-- denied by love policy
--jail.io.lines        = io.lines	-- denied by love policy
--jail.io.popen        = io.popen	-- RISK
--jail.io.tmpfile      = io.tmpfile	-- denied by love policy


-----------------------------
--[[ math ]]-----------------
-----------------------------

jail.math   = math			-- ok


-----------------------------
--[[ string ]]---------------
-----------------------------

jail.string = string			-- ok


-----------------------------
--[[ table ]]----------------
-----------------------------

jail.table  = table			-- ok


-----------------------------
--[[ coroutine ]]------------
-----------------------------

--coroutine				-- seems not ok


-----------------------------
--[[ debug ]]----------------
-----------------------------

--debug					-- seems not ok


-----------------------------
--[[ file ]]-----------------
-----------------------------

--file					-- denied by love policy


-----------------------------
--[[ package ]]--------------
-----------------------------

jail.package		= {}
jail.package.preload    = {}

-- TODO: follow standard behavior ? LUA_PATH/os.getenv(LUA_PATH)/default ?
jail.package.path	= "?;?.lua"

local function jail_package_loadlib(modname)
	-- load it out-of-the-jail
	local modobj = require(modname)
	return modobj
end

jail.package.cpath		= ""	-- RISK: for C lib, disable
jail.package.loadlib = nil              -- RISK: for C lib, disable

local function jail_package_seeall(mod)
	local mt = {__index = jail._G}
	local m = {}
	-- register module m into modules ?
	-- TODO: what exactly doing ?
	return setmetatable(mt, m)
end
jail.package.seeall	= jail_package_seeall	-- RISK: attached to global env ?!

-- package.loaders :
-- 1.  package.preload (ignore)
-- 2.  package.path    (wanted)
-- 3.  package.cpath   (required?)
-- 4.  all-in-one      (ignore)

-- 1. emulate the first loader
local function jail_package_loaders_1(name)
	return string.format("\n\tno field package.preload['%s']", name)
end
-- 2. emulate the second loader
local function jail_package_loaders_2(name)
	local e = ""
	local path = package.path
	name = name:gsub("%.", '/')
	if path and path ~= "" then
		for p in string.gmatch(package.path, "[^;]+") do
			local here = p:gsub("%?", name)
			local f, _e = loadfile(here)
			if f then
				return f
			end
			e=e.."\n\tno file '"..here.."'"
		end
	end
	return e
end
-- 3. emulate the third loader
local function jail_package_loaders_3(name)
	return jail_package_loadlib(name)
end
-- 4. emulate the fouth loader
local function jail_package_loaders_4(name)
	return
end
-- setup the loaders
jail.package.loaders = {
	jail_package_loaders_1,
	jail_package_loaders_2,
	jail_package_loaders_3,
	jail_package_loaders_4,
}

jail.package.loaded = {
	string		= jail.string,
	debug		= jail.debug,
	package		= jail.package,
	_G		= jail._G,
	io		= jail.io,
	os		= jail.os,
	table		= jail.table,
	math		= jail.math,
	coroutine	= jail.coroutine,
}


-----------------------------
---- jail facilities --------
-----------------------------

--[[
local function loadmodule(modname, modobj)
	if jail and jail.package and jail.package.loaded then
		jail.package.loaded[modname] = modobj
		jail[modname] = modobj
	end
end
]]--

local function unloadmodule(modname)
	if jail and jail.package and jail.package.loaded and jail.package.loaded[modname] then
		jail.package.loaded[modname] = nil
		if jail[modname] then
			jail[modname] = nil
		end
	end
end


-- the command line arguments
jail.arg    = arg			-- ok


-- module namespace in local env
local M = {}
M._VERSION = luajailversion -- the module version here
M.jail = jail

M.disable = false
M.unloadmodule = unloadmodule

--function luajail.setupJailFor(f) setfenv(f, jail) end
function M.run(f, ...)
	if not M.disable then
		setfenv(f, jail)
	end
	return f(...)
end

-- module namespace in global env
luajail = M
